//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.


#include "SecondOrderTerm.h"
#include "DOFVector.h"

namespace AMDiS {

  void SecondOrderTerm::lalt(const DimVec<WorldVector<double> >& Lambda,
			     const WorldMatrix<double>& matrix,
			     mtl::dense2D<double>& LALt,
			     bool symm,
			     double factor) const
  {
    int dim = num_rows(LALt);
  
    if (symm) {
      for (int i = 0; i < dim; i++) {
	double val = 0.0;
	for (int k = 0; k < dimOfWorld; k++)
	  for (int l = 0; l < dimOfWorld; l++)
	    val += Lambda[i][k] * matrix[k][l] * Lambda[i][l];
	val *= factor;
	LALt[i][i] += val;
	for (int j = i + 1; j < dim; j++) {
	  val = 0.0;
	  for (int k = 0; k < dimOfWorld; k++)
	    for (int l = 0; l < dimOfWorld; l++)
	      val += Lambda[i][k] * matrix[k][l] * Lambda[j][l];
	  val *= factor;
	  LALt[i][j] += val;
	  LALt[j][i] += val;
	}
      }    
    } else {	
      for (int i = 0; i < dim; i++) {
	for (int j = 0; j < dim; j++) {
	  double val = 0.0;
	  for (int k = 0; k < dimOfWorld; k++)
	    for (int l = 0; l < dimOfWorld; l++) 
	      val += Lambda[i][k] * matrix[k][l] * Lambda[j][l];
	  val *= factor;
	  LALt[i][j] += val;
	}
      }    
    }      
  }


  void SecondOrderTerm::lalt_kl(const DimVec<WorldVector<double> >& Lambda,
				int k, int l,
				mtl::dense2D<double>& LALt,
				double factor)
  {
    int dim = num_rows(LALt);

    for (int i = 0; i < dim; i++)
      for (int j = 0; j < dim; j++)
	LALt[i][j] += factor * Lambda[i][k] * Lambda[j][l];
  }


  // ========== MatrixFct_SOT ==========

  MatrixFct_SOT::MatrixFct_SOT(DOFVectorBase<double> *dv, 
			       AbstractFunction<WorldMatrix<double>, double> *fct,
			       AbstractFunction<WorldVector<double>, WorldMatrix<double> > *div,
			       bool sym)
    : SecondOrderTerm(fct->getDegree()), 
      vec(dv), 
      matrixFct(fct), 
      divFct(div),
      symmetric(sym)
  {
    setSymmetric(symmetric);

    TEST_EXIT(dv)("No vector!\n");

    auxFeSpaces.insert(dv->getFeSpace());
  }

  void MatrixFct_SOT::initElement(const ElInfo* elInfo, 
				  SubAssembler* subAssembler,
				  Quadrature *quad) 
  {
    getVectorAtQPs(vec, elInfo, subAssembler, quad, vecAtQPs);
  }

  void MatrixFct_SOT::getLALt(const ElInfo *elInfo, 
			      vector<mtl::dense2D<double> > &LALt) const
  {
    const DimVec<WorldVector<double> > &grdLambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(LALt.size());

    for (int iq = 0; iq < nPoints; iq++)
      lalt(grdLambda, (*matrixFct)(vecAtQPs[iq]), LALt[iq], symmetric, 1.0);
  }

  void MatrixFct_SOT::eval( int nPoints,
			    const mtl::dense_vector<double>& uhAtQP,
			    const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
			    const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
			    mtl::dense_vector<double>& result,
			    double factor) 
  {
    int dow = Global::getGeo(WORLD);

    for (int iq = 0; iq < nPoints; iq++) {
      double resultQP = 0.0;

      WorldMatrix<double> A = (*matrixFct)(vecAtQPs[iq]);

      if (num_rows(D2UhAtQP) > 0)
	for (int i = 0; i < dow; i++)
	  for (int j = 0; j < dow; j++)
	    resultQP += A[i][j] * D2UhAtQP[iq][j][i];

      if (num_rows(grdUhAtQP) > 0)
	resultQP += (*divFct)(A) * grdUhAtQP[iq];

      result[iq] += resultQP * factor;
    }
  }

  void MatrixFct_SOT::weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
			       std::vector<WorldVector<double> > &result)
  {
    int nPoints = grdUhAtQP.size();
    WorldMatrix<double> A;
    for (int iq = 0; iq < nPoints; iq++) {
      A = (*matrixFct)(vecAtQPs[iq]);
      result[iq] += A * grdUhAtQP[iq];
    }    
  }


  // ========== Matrix_SOT ==========

  void Matrix_SOT::eval(int nPoints,
			const mtl::dense_vector<double>& uhAtQP,
			const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
			const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
			mtl::dense_vector<double>& result,
			double factor) 
  {
    int dow = Global::getGeo(WORLD);
  
    if (num_rows(D2UhAtQP) > 0) {
      for (int iq = 0; iq < nPoints; iq++) {
	double resultQP = 0.0;
	for (int i = 0; i < dow; i++)
	  for (int j = 0; j < dow; j++)
	    resultQP += matrix[i][j] * D2UhAtQP[iq][j][i];	  
	
	result[iq] += resultQP * factor;
      }
    }
  }

  void Matrix_SOT::weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
			    std::vector<WorldVector<double> > &result)
  {
    int nPoints = grdUhAtQP.size();
    for (int iq = 0; iq < nPoints; iq++)
      result[iq] += matrix * grdUhAtQP[iq];
  }


  // ========== VecAtQP_SOT ==========

  VecAtQP_SOT::VecAtQP_SOT(DOFVectorBase<double> *dv, 
			   AbstractFunction<double, double> *af)
    : SecondOrderTerm(af ? af->getDegree() : dv->getFeSpace()->getBasisFcts()->getDegree()), vec(dv), f(af)
  {
    setSymmetric(true);

    TEST_EXIT(dv)("No vector!\n");

    auxFeSpaces.insert(dv->getFeSpace());
  }

  void VecAtQP_SOT::initElement(const ElInfo* elInfo, 
				SubAssembler* subAssembler,
				Quadrature* quad) 
  {
    getVectorAtQPs(vec, elInfo, subAssembler, quad, vecAtQPs);
  }

  void VecAtQP_SOT::initElement(const ElInfo* smallElInfo, 
				const ElInfo* largeElInfo,
				SubAssembler* subAssembler,
				Quadrature* quad) 
  {
    getVectorAtQPs(vec, smallElInfo, largeElInfo, subAssembler, quad, vecAtQPs);
  }

  void VecAtQP_SOT::getLALt(const ElInfo *elInfo, 
			    vector<mtl::dense2D<double> > &LALt) const 
  {
    const DimVec<WorldVector<double> > &grdLambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(LALt.size());

    if (f) {
      for (int iq = 0; iq < nPoints; iq++)
	l1lt(grdLambda, LALt[iq], (*f)(vecAtQPs[iq]));
    } else {
      for (int iq = 0; iq < nPoints; iq++)
	l1lt(grdLambda, LALt[iq], vecAtQPs[iq]);
    }
  }

  void VecAtQP_SOT::eval( int nPoints,
			  const mtl::dense_vector<double>& uhAtQP,
			  const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
			  const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
			  mtl::dense_vector<double>& result,
			  double fac) 
  {
    int dow = Global::getGeo(WORLD);

    if (num_rows(D2UhAtQP) > 0) {
      if (f) {
	for (int iq = 0; iq < nPoints; iq++) {
	  double resultQP = 0.0;
	  for (int i = 0; i < dow; i++)
	    resultQP += D2UhAtQP[iq][i][i];
	  result[iq] += fac * (*f)(vecAtQPs[iq]) * resultQP;
	}
      } else {
	for (int iq = 0; iq < nPoints; iq++) {
	  double resultQP = 0.0;
	  for (int i = 0; i < dow; i++)
	    resultQP += D2UhAtQP[iq][i][i];
	  result[iq] += fac * vecAtQPs[iq] * resultQP;
	}
      }
    }
  }

  void VecAtQP_SOT::weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
			     std::vector<WorldVector<double> > &result) 
  {
    int nPoints = grdUhAtQP.size();
    if (f) {
      for (int iq = 0; iq < nPoints; iq++) {
	double factor = (*f)(vecAtQPs[iq]);
	axpy(factor, grdUhAtQP[iq], result[iq]);
      }
    } else {
      for (int iq = 0; iq < nPoints; iq++) {
	double factor = vecAtQPs[iq];
	axpy(factor, grdUhAtQP[iq], result[iq]);
      }
    }
  }


  // ========== Vec2AtQP_SOT ==========

  Vec2AtQP_SOT::Vec2AtQP_SOT(DOFVectorBase<double> *dv1, DOFVectorBase<double> *dv2, 
			     BinaryAbstractFunction<double, double, double> *af)
    : SecondOrderTerm(af ? af->getDegree() : 2*dv1->getFeSpace()->getBasisFcts()->getDegree()), vec1(dv1), vec2(dv2), f(af)
  {
    setSymmetric(true);

    TEST_EXIT(dv1)("No first vector!\n");
    TEST_EXIT(dv2)("No second vector!\n");

    auxFeSpaces.insert(dv1->getFeSpace());
    auxFeSpaces.insert(dv2->getFeSpace());
  }

  void Vec2AtQP_SOT::initElement(const ElInfo* elInfo, 
				 SubAssembler* subAssembler,
				 Quadrature *quad) 
  {
    getVectorAtQPs(vec1, elInfo, subAssembler, quad, vecAtQPs1);
    getVectorAtQPs(vec2, elInfo, subAssembler, quad, vecAtQPs2);
  }

  void Vec2AtQP_SOT::getLALt(const ElInfo *elInfo, 
			     vector<mtl::dense2D<double> > &LALt) const 
  {
    const DimVec<WorldVector<double> > &grdLambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(LALt.size());

    for (int iq = 0; iq < nPoints; iq++) {
      if (f)
	l1lt(grdLambda, LALt[iq], (*f)(vecAtQPs1[iq], vecAtQPs2[iq]));
      else
	l1lt(grdLambda, LALt[iq], vecAtQPs1[iq] * vecAtQPs2[iq]);
    }
  }

  void Vec2AtQP_SOT::eval(int nPoints,
			  const mtl::dense_vector<double>& uhAtQP,
			  const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
			  const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
			  mtl::dense_vector<double>& result,
			  double fac) 
  {
    int dow = Global::getGeo(WORLD);
    
    if (num_rows(D2UhAtQP) > 0) {
      for (int iq = 0; iq < nPoints; iq++) {
	double factor = 1.0;
	if (f)
	  factor = (*f)(vecAtQPs1[iq], vecAtQPs2[iq]);
	else
	  factor = vecAtQPs1[iq] * vecAtQPs2[iq];
	double resultQP = 0.0;
	for (int i = 0; i < dow; i++)
	  resultQP += D2UhAtQP[iq][i][i];	
	result[iq] += fac * factor * resultQP;
      }
    }
  }
  
  void Vec2AtQP_SOT::weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
			      std::vector<WorldVector<double> > &result)
  {
    int nPoints = grdUhAtQP.size();
    for (int iq = 0; iq < nPoints; iq++) {
      double factor = 1.0;
      if (f)
	factor = (*f)(vecAtQPs1[iq], vecAtQPs2[iq]);
      else
	factor = vecAtQPs1[iq] * vecAtQPs2[iq];
      axpy(factor, grdUhAtQP[iq], result[iq]);      
    }
  }


  // ========== CoordsAtQP_SOT ==========

  void CoordsAtQP_SOT::initElement(const ElInfo* elInfo, 
				   SubAssembler* subAssembler,
				   Quadrature *quad) 
  {
    subAssembler->getCoordsAtQPs(elInfo, quad, coordsAtQPs);
  }

  void CoordsAtQP_SOT::getLALt(const ElInfo *elInfo, 
			       vector<mtl::dense2D<double> > &LALt) const 
  {
    const DimVec<WorldVector<double> > &grdLambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(LALt.size());

    for (int iq = 0; iq < nPoints; iq++)
      l1lt(grdLambda, LALt[iq], (*g)(coordsAtQPs[iq]));
  }

  void CoordsAtQP_SOT::eval(int nPoints,
			    const mtl::dense_vector<double>& uhAtQP,
			    const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
			    const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
			    mtl::dense_vector<double>& result,
			    double f) 
  {
    int dow = Global::getGeo(WORLD);

    if (num_rows(D2UhAtQP) > 0) {
      for (int iq = 0; iq < nPoints; iq++) {
	double factor = (*g)(coordsAtQPs[iq]);
	double resultQP = 0.0;
	for (int i = 0; i < dow; i++)
	  resultQP += D2UhAtQP[iq][i][i];	
	result[iq] += factor * f * resultQP;
      }
    }
  }

  void CoordsAtQP_SOT::weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
				std::vector<WorldVector<double> > &result) 
  {
    FUNCNAME("CoordsAtQP_SOT::weakEval()");
    TEST_EXIT_DBG(grdUhAtQP.size() == num_rows(coordsAtQPs))
      ("Wrong sizes! grdUhAtQP = %d   coordsAtQPs = %d\n", 
       grdUhAtQP.size(), num_rows(coordsAtQPs));

    int nPoints = grdUhAtQP.size();
    for (int iq = 0; iq < nPoints; iq++) {
      double factor = (*g)(coordsAtQPs[iq]);  
      axpy(factor, grdUhAtQP[iq], result[iq]);      
    }
  }


  // ========== MatrixGradient_SOT ==========

  MatrixGradient_SOT::MatrixGradient_SOT(DOFVectorBase<double> *dv,
					 AbstractFunction<WorldMatrix<double>, WorldVector<double> > *af,
					 AbstractFunction<WorldVector<double>, WorldMatrix<double> > *divAf,
					 bool symm) 
    : SecondOrderTerm(af->getDegree()), 
      vec(dv), 
      f(af), 
      divFct(divAf),
      symmetric(symm)
  {
    setSymmetric(symmetric);

    TEST_EXIT(dv)("No vector!\n");

    auxFeSpaces.insert(dv->getFeSpace());
  }

  void MatrixGradient_SOT::initElement(const ElInfo* elInfo, 
				       SubAssembler* subAssembler,
				       Quadrature *quad) 
  {
    getGradientsAtQPs(vec, elInfo, subAssembler, quad, gradAtQPs);
  }

  void MatrixGradient_SOT::initElement(const ElInfo* smallElInfo, 
				       const ElInfo* largeElInfo,
				       SubAssembler* subAssembler,
				       Quadrature *quad) 
  {
    getGradientsAtQPs(vec, smallElInfo, largeElInfo, subAssembler, quad, gradAtQPs);
  }

  void MatrixGradient_SOT::getLALt(const ElInfo *elInfo, 
				   vector<mtl::dense2D<double> > &LALt) const 
  {
    const DimVec<WorldVector<double> > &grdLambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(LALt.size());

    for (int iq = 0; iq < nPoints; iq++)
      lalt(grdLambda, (*f)(gradAtQPs[iq]), LALt[iq], symmetric, 1.0);
  }

  void MatrixGradient_SOT::eval(int nPoints,
				const mtl::dense_vector<double>& uhAtQP,
				const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
				const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
				mtl::dense_vector<double>& result,
				double factor) 
  {
    int dow = Global::getGeo(WORLD);

    for (int iq = 0; iq < nPoints; iq++) {
      double resultQP = 0.0;

      WorldMatrix<double> A = (*f)(gradAtQPs[iq]);

      if (num_rows(D2UhAtQP) > 0)
	for (int i = 0; i < dow; i++)
	  for (int j = 0; j < dow; j++)
	    resultQP += A[i][j] * D2UhAtQP[iq][j][i];

      if (num_rows(grdUhAtQP) > 0)
	resultQP += (*divFct)(A) * grdUhAtQP[iq];      

      result[iq] += resultQP * factor;
    }
  }


  void MatrixGradient_SOT::weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
				    std::vector<WorldVector<double> > &result)
  {
    FUNCNAME("MatrixGradient_SOT::weakEval()");

    TEST_EXIT_DBG(f)("No function f!\n");
    TEST_EXIT_DBG(num_rows(gradAtQPs))
      ("Operator was not initialized correctly!\n");

    int nPoints = grdUhAtQP.size();
    for (int iq = 0; iq < nPoints; iq++) {
      tmpMat = (*f)(gradAtQPs[iq]);
      result[iq] += tmpMat * grdUhAtQP[iq];    
    }
  }


  // ========== FctGradient_SOT ==========

  FctGradient_SOT::FctGradient_SOT(DOFVectorBase<double> *dv,
				   AbstractFunction<double, WorldVector<double> > *af)
    : SecondOrderTerm(af->getDegree()), vec(dv), f(af)
  {
    TEST_EXIT(dv)("No vector!\n");

    auxFeSpaces.insert(dv->getFeSpace());
  }

  void FctGradient_SOT::initElement(const ElInfo* elInfo, 
				    SubAssembler* subAssembler,
				    Quadrature *quad) 
  {
    getGradientsAtQPs(vec, elInfo, subAssembler, quad, gradAtQPs);
  }

  void FctGradient_SOT::getLALt(const ElInfo *elInfo, 
				vector<mtl::dense2D<double> > &LALt) const 
  {
    const DimVec<WorldVector<double> > &grdLambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(LALt.size());

    for (int iq = 0; iq < nPoints; iq++)
      l1lt(grdLambda, LALt[iq], (*f)(gradAtQPs[iq]));
  }

  void FctGradient_SOT::eval( int nPoints,
			      const mtl::dense_vector<double>& uhAtQP,
			      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
			      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
			      mtl::dense_vector<double>& result,
			      double fac) 
  {
    int dow = Global::getGeo(WORLD);

    if (num_rows(D2UhAtQP) > 0) {
      for (int iq = 0; iq < nPoints; iq++) {
	double factor = (*f)(gradAtQPs[iq]);
	double resultQP = 0.0;
	for (int i = 0; i < dow; i++)
	  resultQP += D2UhAtQP[iq][i][i];	
	result[iq] += fac * factor * resultQP;
      }
    }
  }

  void FctGradient_SOT::weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
				 std::vector<WorldVector<double> > &result) 
  {
    int nPoints = grdUhAtQP.size();
    for (int iq = 0; iq < nPoints; iq++) {
      double factor = (*f)(gradAtQPs[iq]);
      axpy(factor, grdUhAtQP[iq], result[iq]);      
    }
  }


  // ========== VecMatrixGradientAtQP_SOT ==========
  
  VecMatrixGradientAtQP_SOT::VecMatrixGradientAtQP_SOT(DOFVectorBase<double> *dv,
						       BinaryAbstractFunction<WorldMatrix<double>, double, WorldVector<double> > *af,
						       AbstractFunction<WorldVector<double>, WorldMatrix<double> > *divAf,
						       bool symm)
    : SecondOrderTerm(af->getDegree()), vec(dv), f(af), 
      divFct(divAf), symmetric(symm)
  {
    setSymmetric(symmetric);

    TEST_EXIT(dv)("No vector!\n");

    auxFeSpaces.insert(dv->getFeSpace());
  }

  void VecMatrixGradientAtQP_SOT::initElement(const ElInfo* elInfo, 
					      SubAssembler* subAssembler,
					      Quadrature *quad) 
  {
    getVectorAtQPs(vec, elInfo, subAssembler, quad, vecAtQPs);
    getGradientsAtQPs(vec, elInfo, subAssembler, quad, gradAtQPs);
  }

  void VecMatrixGradientAtQP_SOT::getLALt(const ElInfo *elInfo,
					  vector<mtl::dense2D<double> > &LALt) const 
  {
    const DimVec<WorldVector<double> > &grdLambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(LALt.size());

    for (int iq = 0; iq < nPoints; iq++)
      lalt(grdLambda, (*f)(vecAtQPs[iq], gradAtQPs[iq]), LALt[iq], symmetric, 1.0);
  }

  void VecMatrixGradientAtQP_SOT::eval( int nPoints,
					const mtl::dense_vector<double>& uhAtQP,
					const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
					const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
					mtl::dense_vector<double>& result,
				        double factor) 
  {
    int dow = Global::getGeo(WORLD);
    
    for (int iq = 0; iq < nPoints; iq++) {
      double resultQP = 0.0;
      
      WorldMatrix<double> A = (*f)(vecAtQPs[iq], gradAtQPs[iq]);
      
      if (num_rows(D2UhAtQP) > 0)
	for (int i = 0; i < dow; i++)
	  for (int j = 0; j < dow; j++)
	    resultQP += A[i][j] * D2UhAtQP[iq][j][i];
      
      if (num_rows(grdUhAtQP) > 0)
	resultQP += (*divFct)(A) * grdUhAtQP[iq];      
      
      result[iq] += resultQP * factor;
    }
  }

  void VecMatrixGradientAtQP_SOT::weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
					   std::vector<WorldVector<double> > &result) 
  {
    int nPoints = grdUhAtQP.size();
    WorldMatrix<double> A;
    for (int iq = 0; iq < nPoints; iq++) {
      A = (*f)(vecAtQPs[iq], gradAtQPs[iq]);
      result[iq] += A * grdUhAtQP[iq];      
    }
  }


  // ========== VecGradCoordsAtQP_SOT ==========

  VecGradCoordsAtQP_SOT::VecGradCoordsAtQP_SOT(DOFVectorBase<double> *dv,
					       TertiaryAbstractFunction<double, double,
					       WorldVector<double>, WorldVector<double> > *af)
    : SecondOrderTerm(af->getDegree()), vec(dv), f(af)
  {
    TEST_EXIT(dv)("No vector!\n");

    auxFeSpaces.insert(dv->getFeSpace());
  }

  void VecGradCoordsAtQP_SOT::initElement(const ElInfo* elInfo, 
					  SubAssembler* subAssembler,
					  Quadrature *quad) 
  {
    getVectorAtQPs(vec, elInfo, subAssembler, quad, vecAtQPs);
    getGradientsAtQPs(vec, elInfo, subAssembler, quad, gradAtQPs);
    subAssembler->getCoordsAtQPs(elInfo, quad, coordsAtQPs);  
  }

  void VecGradCoordsAtQP_SOT::getLALt(const ElInfo *elInfo, 
				      vector<mtl::dense2D<double> > &LALt) const 
  {
    const DimVec<WorldVector<double> > &grdLambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(LALt.size());

    for (int iq = 0; iq < nPoints; iq++)
      l1lt(grdLambda, LALt[iq], (*f)(vecAtQPs[iq], gradAtQPs[iq], coordsAtQPs[iq]));
  }

  void VecGradCoordsAtQP_SOT::eval( int nPoints,
				    const mtl::dense_vector<double>& uhAtQP,
				    const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
				    const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
				    mtl::dense_vector<double>& result,
				    double fac) 
  {
    int  dow = Global::getGeo(WORLD);

    if (num_rows(D2UhAtQP) > 0) {
      for (int iq = 0; iq < nPoints; iq++) {
	double factor = (*f)(vecAtQPs[iq], gradAtQPs[iq], coordsAtQPs[iq]);
	double resultQP = 0.0;
	for (int i = 0; i < dow; i++)
	  resultQP += D2UhAtQP[iq][i][i];	
	result[iq] += fac * factor * resultQP;
      }
    }
  }

  void VecGradCoordsAtQP_SOT::weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
				       std::vector<WorldVector<double> > &result) 
  {
    int nPoints = grdUhAtQP.size();
    for (int iq = 0; iq < nPoints; iq++) {
      double factor = (*f)(vecAtQPs[iq], gradAtQPs[iq], coordsAtQPs[iq]);
      axpy(factor, grdUhAtQP[iq], result[iq]);    
    }
  }


  // ========== VecAndCoordsAtQP_SOT ==========

  VecAndCoordsAtQP_SOT::VecAndCoordsAtQP_SOT(DOFVectorBase<double> *dv, 
					     BinaryAbstractFunction<double, double, WorldVector<double> > *af)
    : SecondOrderTerm(af->getDegree()), vec(dv), f(af)
  {
    setSymmetric(true);

    TEST_EXIT(dv)("No vector!\n");

    auxFeSpaces.insert(dv->getFeSpace());
  }

  void VecAndCoordsAtQP_SOT::initElement(const ElInfo* elInfo, 
					 SubAssembler* subAssembler,
					 Quadrature *quad) 
  {
    getVectorAtQPs(vec, elInfo, subAssembler, quad, vecAtQPs);
    subAssembler->getCoordsAtQPs(elInfo, quad, coordsAtQPs);
  }

  void VecAndCoordsAtQP_SOT::getLALt(const ElInfo *elInfo, 
				     vector<mtl::dense2D<double> > &LALt) const 
  { 
    const DimVec<WorldVector<double> > &grdLambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(LALt.size());

    for (int iq = 0; iq < nPoints; iq++)
      l1lt(grdLambda, LALt[iq], (*f)(vecAtQPs[iq], coordsAtQPs[iq]));
  }

  void VecAndCoordsAtQP_SOT::eval(int nPoints,
				  const mtl::dense_vector<double>& uhAtQP,
				  const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
				  const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
				  mtl::dense_vector<double>& result,
				  double fac) 
  {
    int dow = Global::getGeo(WORLD);

    if (num_rows(D2UhAtQP) > 0) {
      for (int iq = 0; iq < nPoints; iq++) {
	double factor = (*f)(vecAtQPs[iq], coordsAtQPs[iq]);
	double resultQP = 0.0;
	for (int i = 0; i < dow; i++)
	  resultQP += D2UhAtQP[iq][i][i];	
	result[iq] += fac * factor * resultQP;
      }
    }
  }

  void VecAndCoordsAtQP_SOT::weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
				      std::vector<WorldVector<double> > &result) 
  {
    int nPoints = grdUhAtQP.size();
    for (int iq = 0; iq < nPoints; iq++) {
      double factor = (*f)(vecAtQPs[iq], coordsAtQPs[iq]);  
      axpy(factor, grdUhAtQP[iq], result[iq]);    
    }
  }


  // ========== MatrixGradientAndCoords_SOT ==========
  
  MatrixGradientAndCoords_SOT::MatrixGradientAndCoords_SOT(DOFVectorBase<double> *dv,
							   BinaryAbstractFunction<WorldMatrix<double>,
							   WorldVector<double>, WorldVector<double> > *af,
							   AbstractFunction<WorldVector<double>,
							   WorldMatrix<double> > *divAf,
							   bool symm) 
    : SecondOrderTerm(af->getDegree()), vec(dv), f(af), divFct(divAf), symmetric(symm)
  {
    setSymmetric(symmetric);

    TEST_EXIT(dv)("No vector!\n");

    auxFeSpaces.insert(dv->getFeSpace());
  }

  void MatrixGradientAndCoords_SOT::initElement(const ElInfo* elInfo, 
						SubAssembler* subAssembler,
						Quadrature *quad) 
  {
    getGradientsAtQPs(vec, elInfo, subAssembler, quad, gradAtQPs);
    subAssembler->getCoordsAtQPs(elInfo, quad, coordsAtQPs);  
  }

  void MatrixGradientAndCoords_SOT::getLALt(const ElInfo *elInfo, 
					    vector<mtl::dense2D<double> > &LALt) const 
  {
    const DimVec<WorldVector<double> > &grdLambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(LALt.size());

    for (int iq = 0; iq < nPoints; iq++)
      lalt(grdLambda, (*f)(gradAtQPs[iq], coordsAtQPs[iq]), LALt[iq], symmetric, 1.0);
  }

  void MatrixGradientAndCoords_SOT::eval( int nPoints,
					  const mtl::dense_vector<double>& uhAtQP,
					  const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
					  const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
					  mtl::dense_vector<double>& result,
					  double factor) 
  {
    int dow = Global::getGeo(WORLD);

    for (int iq = 0; iq < nPoints; iq++) {
      double resultQP = 0.0;

      WorldMatrix<double> A = (*f)(gradAtQPs[iq], coordsAtQPs[iq]);

      if (num_rows(D2UhAtQP) > 0)
	for (int i = 0; i < dow; i++)
	  for (int j = 0; j < dow; j++)
	    resultQP += A[i][j] * D2UhAtQP[iq][j][i];

      if (num_rows(grdUhAtQP) > 0)
	resultQP += (*divFct)(A) * grdUhAtQP[iq];
      
      result[iq] += resultQP * factor;
    }
  }

  void MatrixGradientAndCoords_SOT::weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
					     std::vector<WorldVector<double> > &result) 
  {
    int nPoints = grdUhAtQP.size();
    WorldMatrix<double> A;
    for (int iq = 0; iq < nPoints; iq++) {
      A = (*f)(gradAtQPs[iq], coordsAtQPs[iq]);
      result[iq] += A * grdUhAtQP[iq];      
    }
  }

  
  // ========== MatrixVec2_SOT ==========

  MatrixVec2_SOT::MatrixVec2_SOT(DOFVectorBase<double> *dv1, DOFVectorBase<double> *dv2, 
			       BinaryAbstractFunction<double, double, double> *f,
			       WorldMatrix<double> Af,
			       bool sym)
    : SecondOrderTerm(f->getDegree()), 
      vec1(dv1),
      vec2(dv2), 
      fct(f), 
      A(Af),
      symmetric(sym)
  {
    FUNCNAME("MatrixVec2_SOT::MatrixVec2_SOT()");

    setSymmetric(symmetric);

    TEST_EXIT_DBG(dv1)("No vector!\n");
    TEST_EXIT_DBG(dv2)("No vector!\n");
    
    auxFeSpaces.insert(dv1->getFeSpace());
    auxFeSpaces.insert(dv2->getFeSpace());
  }

  void MatrixVec2_SOT::initElement(const ElInfo* elInfo, 
				  SubAssembler* subAssembler,
				  Quadrature *quad) 
  {
    getVectorAtQPs(vec1, elInfo, subAssembler, quad, vec1AtQPs);
    getVectorAtQPs(vec2, elInfo, subAssembler, quad, vec2AtQPs);
  }
  
  void MatrixVec2_SOT::getLALt(const ElInfo *elInfo, 
			       vector<mtl::dense2D<double> > &LALt) const 
  {      
    const DimVec<WorldVector<double> > &grdLambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(LALt.size());

    for (int iq = 0; iq < nPoints; iq++)
      lalt(grdLambda, A, LALt[iq], symmetric, (*fct)(vec1AtQPs[iq], vec2AtQPs[iq]));
  }

  void MatrixVec2_SOT::eval(int nPoints,
			    const mtl::dense_vector<double>& uhAtQP,
			    const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
			    const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
			    mtl::dense_vector<double>& result,
			    double factor) 
  {
    int dow = Global::getGeo(WORLD);
    
    for (int iq = 0; iq < nPoints; iq++) {
      double resultQP = 0.0;

      if (num_rows(D2UhAtQP) > 0)
	for (int i = 0; i < dow; i++)
	  for (int j = 0; j < dow; j++)
	    resultQP += A[i][j] * D2UhAtQP[iq][j][i];
    
      result[iq] += resultQP * factor * (*fct)(vec1AtQPs[iq], vec2AtQPs[iq]);
    }
  }

  void MatrixVec2_SOT::weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
				std::vector<WorldVector<double> > &result)
  {
    int nPoints = grdUhAtQP.size();
    for (int iq = 0; iq < nPoints; iq++)
      result[iq] += A * grdUhAtQP[iq] * (*fct)(vec1AtQPs[iq], vec2AtQPs[iq]);
  }

  
  // ========== General_SOT ==========

  General_SOT::General_SOT(std::vector<DOFVectorBase<double>*> vecs,
			   std::vector<DOFVectorBase<double>*> grads,
			   TertiaryAbstractFunction<WorldMatrix<double>, 
  			     WorldVector<double>,
			     std::vector<double>, 
			     std::vector<WorldVector<double> > > *f,
			   AbstractFunction<WorldVector<double>, 
			     WorldMatrix<double> > *divFct,
			     bool symmetric)
      : SecondOrderTerm(f->getDegree()),
	vecs_(vecs),
	grads_(grads),
	f_(f),
	divFct_(divFct),
	symmetric_(symmetric)
  {
    vecsAtQPs.resize(vecs_.size());
    gradsAtQPs_.resize(grads_.size());

    for (int i = 0; i < static_cast<int>(vecs.size()); i++) {
      TEST_EXIT(vecs[i])("One vector is NULL!\n");

      auxFeSpaces.insert(vecs[i]->getFeSpace());
    }   

    for (int i = 0; i < static_cast<int>(grads.size()); i++) {
      TEST_EXIT(grads[i])("One gradient vector is NULL!\n");

      auxFeSpaces.insert(grads[i]->getFeSpace());
    }
  }

  void General_SOT::initElement(const ElInfo* elInfo, 
				SubAssembler* subAssembler,
				Quadrature *quad)
  {
    int nVecs = static_cast<int>(vecs_.size());
    int nGrads = static_cast<int>(grads_.size());

    subAssembler->getCoordsAtQPs(elInfo, quad, coordsAtQPs);

    for (int i = 0; i < nVecs; i++)
      getVectorAtQPs(vecs_[i], elInfo, subAssembler, quad, vecsAtQPs[i]);
    for (int i = 0; i < nGrads; i++)
      getGradientsAtQPs(grads_[i], elInfo, subAssembler, quad, gradsAtQPs_[i]);
  }

  void General_SOT::getLALt(const ElInfo *elInfo, 
			    vector<mtl::dense2D<double> > &LALt) const
  {
    int nVecs = static_cast<int>(vecs_.size());
    int nGrads = static_cast<int>(grads_.size());
    const int nPoints = static_cast<int>(LALt.size());

    vector<double> vecsArg(nVecs);
    vector<WorldVector<double> > gradsArg(nGrads);

    const DimVec<WorldVector<double> > &grdLambda = elInfo->getGrdLambda();

    for (int iq = 0; iq < nPoints; iq++) {
      for (int i = 0; i < nVecs; i++)
	vecsArg[i] = vecsAtQPs[i][iq];      
      for (int i = 0; i < nGrads; i++) 
	gradsArg[i] = gradsAtQPs_[i][iq];      

      lalt(grdLambda, (*f_)(coordsAtQPs[iq], vecsArg, gradsArg), 
	   LALt[iq], symmetric_, 1.0);
    }
  }

  void General_SOT::eval( int nPoints,
			  const mtl::dense_vector<double>& uhAtQP,
			  const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
			  const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
			  mtl::dense_vector<double>& result,
			  double factor) 
  {
    int dow = Global::getGeo(WORLD);
    int nVecs = static_cast<int>(vecs_.size());
    int nGrads = static_cast<int>(grads_.size());

    std::vector<double> vecsArg(nVecs);
    std::vector<WorldVector<double> > gradsArg(nGrads);

    for (int iq = 0; iq < nPoints; iq++) {
      double resultQP = 0.0;

      for (int i = 0; i < nVecs; i++)
	vecsArg[i] = vecsAtQPs[i][iq];
      for (int i = 0; i < nGrads; i++)
	gradsArg[i] = gradsAtQPs_[i][iq];

      WorldMatrix<double> A = (*f_)(coordsAtQPs[iq], vecsArg, gradsArg);

      if (num_rows(D2UhAtQP) > 0) {
	for (int i = 0; i < dow; i++)
	  for (int j = 0; j < dow; j++)
	    resultQP += A[i][j] * D2UhAtQP[iq][j][i];
      }

      if (num_rows(grdUhAtQP) > 0)
	resultQP += (*divFct_)(A) * grdUhAtQP[iq];      

      result[iq] += resultQP * factor;
    }
  }

  void General_SOT::weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
			     std::vector<WorldVector<double> > &result) 
  {
    int nPoints = grdUhAtQP.size();
    int nVecs = static_cast<int>(vecs_.size());
    int nGrads = static_cast<int>(grads_.size());

    std::vector<double> vecsArg(nVecs);
    std::vector<WorldVector<double> > gradsArg(nGrads);

    WorldMatrix<double> A;
    for (int iq = 0; iq < nPoints; iq++) {
      for (int i = 0; i < nVecs; i++)
	vecsArg[i] = vecsAtQPs[i][iq];	
      for (int i = 0; i < nGrads; i++)
	gradsArg[i] = gradsAtQPs_[i][iq];
      
      A = (*f_)(coordsAtQPs[iq], vecsArg, gradsArg);
      result[iq] += A * grdUhAtQP[iq];      
    }
  }


  // ========== GeneralParametric_SOT ==========

  GeneralParametric_SOT::GeneralParametric_SOT(std::vector<DOFVectorBase<double>*> vecs,
					       std::vector<DOFVectorBase<double>*> grads,
					       QuartAbstractFunction<WorldMatrix<double>, 
					         WorldVector<double>,
					         WorldVector<double>,
					         std::vector<double>, 
					         std::vector<WorldVector<double> > > *f,
					       AbstractFunction<WorldVector<double>, 
					         WorldMatrix<double> > *divFct,
					       bool symmetric)
    : SecondOrderTerm(f->getDegree()),
      vecs_(vecs),
      grads_(grads),
      f_(f),
      divFct_(divFct),
      symmetric_(symmetric)
  {
    vecsAtQPs.resize(vecs_.size());
    gradsAtQPs_.resize(grads_.size());

    for (int i = 0; i < static_cast<int>(vecs.size()); i++) {
      TEST_EXIT(vecs[i])("One vector is NULL!\n");

      auxFeSpaces.insert(vecs[i]->getFeSpace());
    }   

    for (int i = 0; i < static_cast<int>(grads.size()); i++) {
      TEST_EXIT(grads[i])("One gradient vector is NULL!\n");

      auxFeSpaces.insert(grads[i]->getFeSpace());
    }
  }

  void GeneralParametric_SOT::initElement(const ElInfo* elInfo, 
					  SubAssembler* subAssembler,
					  Quadrature *quad)
  {
    int nVecs = static_cast<int>(vecs_.size());
    int nGrads = static_cast<int>(grads_.size());

    elInfo->getElementNormal(elementNormal_);
    subAssembler->getCoordsAtQPs(elInfo, quad, coordsAtQPs);

    for (int i = 0; i < nVecs; i++)
      getVectorAtQPs(vecs_[i], elInfo, subAssembler, quad, vecsAtQPs[i]);
    for (int i = 0; i < nGrads; i++)
      getGradientsAtQPs(grads_[i], elInfo, subAssembler, quad, gradsAtQPs_[i]);
  }

  void GeneralParametric_SOT::getLALt(const ElInfo *elInfo, 
				      vector<mtl::dense2D<double> > &LALt) const
  {
    int nVecs = static_cast<int>(vecs_.size());
    int nGrads = static_cast<int>(grads_.size());
    const int nPoints = static_cast<int>(LALt.size());

    vector<double> vecsArg(nVecs);
    vector<WorldVector<double> > gradsArg(nGrads);

    const DimVec<WorldVector<double> > &grdLambda = elInfo->getGrdLambda();

    for (int iq = 0; iq < nPoints; iq++) {
      for (int i = 0; i < nVecs; i++)
	vecsArg[i] = vecsAtQPs[i][iq];      
      for (int i = 0; i < nGrads; i++)
	gradsArg[i] = gradsAtQPs_[i][iq];      
      lalt(grdLambda, (*f_)(coordsAtQPs[iq], elementNormal_, vecsArg, gradsArg), 
	   LALt[iq], symmetric_, 1.0);
    }
  }

  void GeneralParametric_SOT::eval( int nPoints,
				    const mtl::dense_vector<double>& uhAtQP,
				    const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
				    const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
				    mtl::dense_vector<double>& result,
				    double factor) 
  {
    int dow = Global::getGeo(WORLD);
    int nVecs = static_cast<int>(vecs_.size());
    int nGrads = static_cast<int>(grads_.size());

    std::vector<double> vecsArg(nVecs);
    std::vector<WorldVector<double> > gradsArg(nGrads);

    for (int iq = 0; iq < nPoints; iq++) {
      double resultQP = 0.0;

      for (int i = 0; i < nVecs; i++)
	vecsArg[i] = vecsAtQPs[i][iq];      
      for (int i = 0; i < nGrads; i++)
	gradsArg[i] = gradsAtQPs_[i][iq];      

      WorldMatrix<double> A = 
	(*f_)(coordsAtQPs[iq],  elementNormal_, vecsArg, gradsArg);

      if (num_rows(D2UhAtQP) > 0) {
	for (int i = 0; i < dow; i++)
	  for (int j = 0; j < dow; j++)
	    resultQP += A[i][j] * D2UhAtQP[iq][j][i];
      }

      if (num_rows(grdUhAtQP) > 0)
	resultQP += (*divFct_)(A) * grdUhAtQP[iq];      

      result[iq] += resultQP * factor;
    }
  }

  void GeneralParametric_SOT::weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
				       std::vector<WorldVector<double> > &result) 
  {
    int nPoints = grdUhAtQP.size();
    int nVecs = static_cast<int>(vecs_.size());
    int nGrads = static_cast<int>(grads_.size());

    std::vector<double> vecsArg(nVecs);
    std::vector<WorldVector<double> > gradsArg(nGrads);

    WorldMatrix<double> A;
    for (int iq = 0; iq < nPoints; iq++) {
      for (int i = 0; i < nVecs; i++)
	vecsArg[i] = vecsAtQPs[i][iq];	
      for (int i = 0; i < nGrads; i++)
	gradsArg[i] = gradsAtQPs_[i][iq];
      
      A = (*f_)(coordsAtQPs[iq],  elementNormal_, vecsArg, gradsArg);
      result[iq] += A * grdUhAtQP[iq];    
    }
  }


  // ========== VecAndGradAtQP_SOT ==========

  VecAndGradAtQP_SOT::VecAndGradAtQP_SOT(DOFVectorBase<double> *dv,
					 BinaryAbstractFunction<double, double, WorldVector<double> > *af)
    : SecondOrderTerm(af->getDegree()), vec(dv), f(af) 
  {
    TEST_EXIT(dv)("No vector!\n");

    auxFeSpaces.insert(dv->getFeSpace());
  }

  void VecAndGradAtQP_SOT::initElement(const ElInfo* elInfo, 
				       SubAssembler* subAssembler,
				       Quadrature *quad) 
  {
    getVectorAtQPs(vec, elInfo, subAssembler, quad, vecAtQPs);
    getGradientsAtQPs(vec, elInfo, subAssembler, quad, gradAtQPs);
  }

  void VecAndGradAtQP_SOT::getLALt(const ElInfo *elInfo, 
				   vector<mtl::dense2D<double> > &LALt) const 
  {
    const DimVec<WorldVector<double> > &grdLambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(LALt.size());

    for (int iq = 0; iq < nPoints; iq++)
      l1lt(grdLambda, LALt[iq], (*f)(vecAtQPs[iq], gradAtQPs[iq]));
  }

  void VecAndGradAtQP_SOT::eval(int nPoints,
				const mtl::dense_vector<double>& uhAtQP,
				const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
				const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
				mtl::dense_vector<double>& result,
				double fac) 
  {
    int dow = Global::getGeo(WORLD);

    if (num_rows(D2UhAtQP) > 0) {
      for (int iq = 0; iq < nPoints; iq++) {
	double factor = (*f)(vecAtQPs[iq], gradAtQPs[iq]);
	double resultQP = 0.0;
	for (int i = 0; i < dow; i++)
	  resultQP += D2UhAtQP[iq][i][i];
	
	result[iq] += fac * resultQP * factor;
      }
    }
  }

  void VecAndGradAtQP_SOT::weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
				    std::vector<WorldVector<double> > &result) 
  {
    int nPoints = grdUhAtQP.size();
    for (int iq = 0; iq < nPoints; iq++) {
      double factor = (*f)(vecAtQPs[iq], gradAtQPs[iq]);  
      axpy(factor, grdUhAtQP[iq], result[iq]);      
    }
  }

  
  // ========== VecGrad_SOT ==========

  void VecGrad_SOT::initElement(const ElInfo* elInfo, 
				SubAssembler* subAssembler,
				Quadrature *quad) 
  {
    subAssembler->getVectorAtQPs(vec1, elInfo, quad, vecAtQPs);
    subAssembler->getGradientsAtQPs(vec2, elInfo, quad, gradAtQPs);
  }

  void VecGrad_SOT::getLALt(const ElInfo *elInfo,
			    vector<mtl::dense2D<double> > &LALt) const 
  {
    const DimVec<WorldVector<double> > &grdLambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(LALt.size());

    for (int iq = 0; iq < nPoints; iq++)
      l1lt(grdLambda, LALt[iq], (*f)(vecAtQPs[iq], gradAtQPs[iq]));    
  }

  void VecGrad_SOT::eval( int nPoints,
			  const mtl::dense_vector<double>& uhAtQP,
			  const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
			  const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
			  mtl::dense_vector<double>& result,
			  double fac) 
  {
    int dow = Global::getGeo(WORLD);

    if (num_rows(D2UhAtQP) > 0) {
      for (int iq = 0; iq < nPoints; iq++) {
	double factor = (*f)(vecAtQPs[iq], gradAtQPs[iq]);
	double resultQP = 0.0;
	for (int i = 0; i < dow; i++)
	  resultQP += D2UhAtQP[iq][i][i];	
	result[iq] += fac * resultQP * factor;
      }
    }
  }

  void VecGrad_SOT::weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
			     std::vector<WorldVector<double> > &result) 
  {
    int nPoints = grdUhAtQP.size();
    for (int iq = 0; iq < nPoints; iq++) {
      double factor = (*f)(vecAtQPs[iq], gradAtQPs[iq]);  
      axpy(factor, grdUhAtQP[iq], result[iq]);
    }
  }

  
  // ========== CoordsAtQP_IJ_SOT ==========

  void CoordsAtQP_IJ_SOT::initElement(const ElInfo* elInfo, 
				      SubAssembler* subAssembler,
				      Quadrature *quad)
  {
    subAssembler->getCoordsAtQPs(elInfo, quad, coordsAtQPs);
  }

  void CoordsAtQP_IJ_SOT::getLALt(const ElInfo *elInfo, 
				  vector<mtl::dense2D<double> > &LALt) const
  {
    const DimVec<WorldVector<double> > Lambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(LALt.size());

    for (int iq = 0; iq < nPoints; iq++)
      lalt_kl(Lambda, xi, xj, LALt[iq], (*g)(coordsAtQPs[iq]));
  }

  void CoordsAtQP_IJ_SOT::eval( int nPoints,
				const mtl::dense_vector<double>& uhAtQP,
				const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
				const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
				mtl::dense_vector<double>& result,
			        double f) 
  {
    if (num_rows(D2UhAtQP) > 0) {
      for (int iq = 0; iq < nPoints; iq++) {
	double factor = (*g)(coordsAtQPs[iq]);
	result[iq] += D2UhAtQP[iq][xi][xj] * factor * f;
      }
    }
  }

  void CoordsAtQP_IJ_SOT::weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
				   std::vector<WorldVector<double> > &result) 
  {
    int nPoints = grdUhAtQP.size();
    for (int iq = 0; iq < nPoints; iq++) {
      double factor = (*g)(coordsAtQPs[iq]);  
      result[iq][xi] += grdUhAtQP[iq][xj] * factor;      
    }
  }

  
  // ========== VecAtQP_IJ_SOT ==========
  
  VecAtQP_IJ_SOT::VecAtQP_IJ_SOT(DOFVectorBase<double> *dv, 
				 AbstractFunction<double, double> *af,
				 int x_i, int x_j)
    : SecondOrderTerm(af ? af->getDegree() : dv->getFeSpace()->getBasisFcts()->getDegree()), vec(dv), f(af), xi(x_i), xj(x_j)
  {
    setSymmetric(xi == xj);

    TEST_EXIT(dv)("No vector!\n");

    auxFeSpaces.insert(dv->getFeSpace());
  }

  void VecAtQP_IJ_SOT::initElement(const ElInfo* elInfo, 
				   SubAssembler* subAssembler,
				   Quadrature *quad)
  {
    getVectorAtQPs(vec, elInfo, subAssembler, quad, vecAtQPs);
  }
 
  void VecAtQP_IJ_SOT::getLALt(const ElInfo *elInfo, 
			       vector<mtl::dense2D<double> > &LALt) const
  {
    const DimVec<WorldVector<double> > &grdLambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(LALt.size());

    if (f) {
      for (int iq = 0; iq < nPoints; iq++)
	lalt_kl(grdLambda, xi, xj, LALt[iq], (*f)(vecAtQPs[iq]));
    } else {
      for (int iq = 0; iq < nPoints; iq++)
	lalt_kl(grdLambda, xi, xj, LALt[iq], vecAtQPs[iq]);
    }
  }  

  void VecAtQP_IJ_SOT::eval(int nPoints,
			    const mtl::dense_vector<double>& uhAtQP,
			    const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
			    const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
			    mtl::dense_vector<double>& result,
			    double fac) 
  {
    if (num_rows(D2UhAtQP) > 0) {
      if (f) {
	for (int iq = 0; iq < nPoints; iq++) {
	  double factor = (*f)(vecAtQPs[iq]);
	  result[iq] += D2UhAtQP[iq][xi][xj] * factor * fac;
	}
      } else {
	for (int iq = 0; iq < nPoints; iq++) {
	  double factor = vecAtQPs[iq];
	  result[iq] += D2UhAtQP[iq][xi][xj] * factor * fac;
	}
      }
    }
  }

  void VecAtQP_IJ_SOT::weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
				std::vector<WorldVector<double> > &result) 
  {
    int nPoints = grdUhAtQP.size();
    if (f) {
      for (int iq = 0; iq < nPoints; iq++) {
	double factor = (*f)(vecAtQPs[iq]);
	result[iq][xi] += grdUhAtQP[iq][xj] * factor;
      }
    } else {
      for (int iq = 0; iq < nPoints; iq++) {
	double factor = vecAtQPs[iq];
	result[iq][xi] += grdUhAtQP[iq][xj] * factor;
      }
    }
  }


  // ========== Vec2AtQP_IJ_SOT ==========

  Vec2AtQP_IJ_SOT::Vec2AtQP_IJ_SOT(DOFVectorBase<double> *dv1,
				   DOFVectorBase<double> *dv2,
				   BinaryAbstractFunction<double, double, double> *af,
				   int x_i, int x_j)
    : SecondOrderTerm(af ? af->getDegree() : dv1->getFeSpace()->getBasisFcts()->getDegree()+dv2->getFeSpace()->getBasisFcts()->getDegree()),
      vec1(dv1),
      vec2(dv2),
      f(af),
      xi(x_i), xj(x_j)
  {
    setSymmetric(xi == xj);

    TEST_EXIT(dv1)("No vector 1!\n");
    TEST_EXIT(dv2)("No vector 2!\n");

    auxFeSpaces.insert(dv1->getFeSpace());
    auxFeSpaces.insert(dv2->getFeSpace());
  }

  void Vec2AtQP_IJ_SOT::initElement(const ElInfo* elInfo,
				   SubAssembler* subAssembler,
				   Quadrature *quad)
  {
    getVectorAtQPs(vec1, elInfo, subAssembler, quad, vecAtQPs1);
    getVectorAtQPs(vec2, elInfo, subAssembler, quad, vecAtQPs2);
  }

  void Vec2AtQP_IJ_SOT::getLALt(const ElInfo *elInfo,
			       vector<mtl::dense2D<double> > &LALt) const
  {
    const DimVec<WorldVector<double> > &grdLambda = elInfo->getGrdLambda();
    const int nPoints = static_cast<int>(LALt.size());

    if (f) {
      for (int iq = 0; iq < nPoints; iq++)
	lalt_kl(grdLambda, xi, xj, LALt[iq], (*f)(vecAtQPs1[iq], vecAtQPs2[iq]));
    } else {
      for (int iq = 0; iq < nPoints; iq++)
	lalt_kl(grdLambda, xi, xj, LALt[iq], vecAtQPs1[iq] * vecAtQPs2[iq]);
    }
  }

  void Vec2AtQP_IJ_SOT::eval(int nPoints,
			    const mtl::dense_vector<double>& uhAtQP,
			    const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
			    const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
			    mtl::dense_vector<double>& result,
			    double fac)
  {
    if (num_rows(D2UhAtQP) > 0) {
      if (f) {
	for (int iq = 0; iq < nPoints; iq++) {
	  double factor = (*f)(vecAtQPs1[iq], vecAtQPs2[iq]);
	  result[iq] += D2UhAtQP[iq][xi][xj] * factor * fac;
	}
      } else {
	for (int iq = 0; iq < nPoints; iq++) {
	  double factor = vecAtQPs1[iq] * vecAtQPs2[iq];
	  result[iq] += D2UhAtQP[iq][xi][xj] * factor * fac;
	}
      }
    }
  }

  void Vec2AtQP_IJ_SOT::weakEval(const std::vector<WorldVector<double> > &grdUhAtQP,
				std::vector<WorldVector<double> > &result)
  {
    int nPoints = grdUhAtQP.size();
    if (f) {
      for (int iq = 0; iq < nPoints; iq++) {
	double factor = (*f)(vecAtQPs1[iq], vecAtQPs2[iq]);
	result[iq][xi] += grdUhAtQP[iq][xj] * factor;
      }
    } else {
      for (int iq = 0; iq < nPoints; iq++) {
	double factor = vecAtQPs1[iq] * vecAtQPs2[iq];
	result[iq][xi] += grdUhAtQP[iq][xj] * factor;
      }
    }
  }

}
