//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.


#include <vector>
#include <set>

#include "io/VtkWriter.h"
#include "parallel/PetscProblemStat.h"
#include "parallel/PetscSolver.h"
#include "parallel/MpiHelper.h"
#include "parallel/BddcMlSolver.h"
#include "parallel/PetscSolverNavierStokes.h"

namespace AMDiS {

  using namespace std;


  PetscProblemStat::PetscProblemStat(string nameStr,
				     ProblemIterationInterface *problemIteration)
    : ParallelProblemStatBase(nameStr, problemIteration),
      deleteSolver(true)
  {
    FUNCNAME("PetscProblemStat::PetscProblemStat()");

  }


  PetscProblemStat::PetscProblemStat(string nameStr,
				     PetscSolver *solver)
    : ParallelProblemStatBase(nameStr, NULL),
      petscSolver(solver),
      deleteSolver(false)
  {}


  void PetscProblemStat::initialize(Flag initFlag, 
				    ProblemStatSeq* adoptProblem,
				    Flag adoptFlag)
  {
    ParallelProblemStatBase::initialize(initFlag, adoptProblem, adoptFlag);
    
    string tmp("");
    Parameters::get(getName() + "->solver->petsc prefix", tmp);
    petscSolver->setKspPrefix(tmp);    
    
    meshDistributor->setBoundaryDofRequirement(petscSolver->getBoundaryDofRequirement());

    petscSolver->setMeshDistributor(meshDistributor, 
				    meshDistributor->getMpiComm(),
				    PETSC_COMM_SELF);   
    petscSolver->init(this->getComponentSpaces(), this->getFeSpaces());
  }
  
    
  void PetscProblemStat::createSolver()
  {
    FUNCNAME("createSolver()");
    ParallelProblemStatBase::createSolver();
    
    petscSolver = dynamic_cast<PetscSolver*>(solver);
    TEST_EXIT(petscSolver)("wrong petsc-solver type\n");
  }


  void PetscProblemStat::solve(AdaptInfo *adaptInfo,
			       bool createMatrixData,
			       bool storeMatrixData)
  {
    FUNCNAME("PetscProblemStat::solve()");

    TEST_EXIT(meshDistributor)("Should not happen!\n");

    double wtime = MPI::Wtime();

    if (createMatrixData)
      petscSolver->fillPetscMatrix(systemMatrix);

    petscSolver->fillPetscRhs(rhs);

    INFO(info, 8)("creation of parallel data structures needed %.5f seconds\n", 
		  MPI::Wtime() - wtime);
    wtime = MPI::Wtime();

    petscSolver->solvePetscMatrix(*solution, adaptInfo);   

    INFO(info, 8)("solution of discrete system needed %.5f seconds\n", 
		  MPI::Wtime() - wtime);

    petscSolver->destroyVectorData();

    if (!storeMatrixData)
      petscSolver->destroyMatrixData();
  }

}
