#include "AMDiS.h"
#include "CahnHilliard.h"
#include "NavierStokes_TaylorHood.h"
#include "CahnHilliardNavierStokes.h"
#include "SignedDistFunctors.h"
#include "PhaseFieldConvert.h"

#include "boost/date_time/posix_time/posix_time.hpp"

using namespace AMDiS;
using namespace boost::posix_time;

struct DrivenCavityBC : AbstractFunction<double, WorldVector<double> >
{
  double operator()(const WorldVector<double> &x) const {
    return std::max(0.0, 1.0 - 4.0 * sqr(x[0] - 0.5));
  }
};


class CH_DrivenCavity : public CahnHilliard
{
public:
  CH_DrivenCavity(std::string name_) : CahnHilliard(name_) {}
  
protected:  
  void fillBoundaryConditions()
  { FUNCNAME("CH_DrivenCavity::fillBoundaryConditions()");
    // homogeneouse neumann conditions
  }
};


class NS_DrivenCavity : public NavierStokes_TaylorHood
{
public:
  NS_DrivenCavity(std::string name_) : NavierStokes_TaylorHood(name_) {}
  
protected:
  void fillBoundaryConditions()
  { FUNCNAME("NS_DrivenCavity::fillBoundaryConditions()");
    
    DOFVector<double> *zeroDOF = new DOFVector<double>(getFeSpace(0), "zero");
    zeroDOF->set(0.0);
    size_t dow = Global::getGeo(WORLD);

    /// at rigid wall: no-slip boundary condition
    for (size_t i = 0; i < dow; i++)
      getProblem(0)->addDirichletBC(1, i, i, zeroDOF);

    /// at upper wall: prescribed velocity
    getProblem(0)->addDirichletBC(2, 0, 0, new DrivenCavityBC);
    getProblem(0)->addDirichletBC(2, 1, 1, zeroDOF);
  }
};


int main(int argc, char** argv)
{ FUNCNAME("main");

  AMDiS::init(argc, argv);

  CH_DrivenCavity chProb("ch");
  NS_DrivenCavity nsProb("ns");

  CahnHilliardNavierStokes<CH_DrivenCavity,NS_DrivenCavity> mainProb("main", &chProb, &nsProb);

  // Adapt-Infos
  AdaptInfo adaptInfo("adapt", mainProb.getNumComponents());
  mainProb.initialize(&adaptInfo);

  // adaption loop - solve ch-prob and ns-prob
  AdaptInstationary adaptInstat("adapt", mainProb, adaptInfo, mainProb, adaptInfo);

  ptime start_time = microsec_clock::local_time();
  int error_code = adaptInstat.adapt(); 
  time_duration td = microsec_clock::local_time()-start_time;

  MSG("elapsed time= %d sec\n", td.total_seconds());

  AMDiS::finalize();

  return error_code;
};
