#include "AMDiS.h"
#include "BaseProblem.h"

#include "boost/date_time/posix_time/posix_time.hpp"

using namespace AMDiS;
using namespace boost::posix_time;

// ===========================================================================
// ===== function definitions ================================================
// ===========================================================================

/// Dirichlet boundary function
class G : public AbstractFunction<double, WorldVector<double> >
{
public:

  /// Implementation of AbstractFunction::operator().
  double operator()(const WorldVector<double>& x) const 
  {
    return exp(-10.0 * (x * x));
  }
};

/// RHS function
class F : public AbstractFunction<double, WorldVector<double> >
{
public:

  F(int degree) : AbstractFunction<double, WorldVector<double> >(degree) {}

  /// Implementation of AbstractFunction::operator().
  double operator()(const WorldVector<double>& x) const 
  {
    int dow = Global::getGeo(WORLD);
    double r2 = (x * x);
    double ux = exp(-10.0 * r2);
    return -(400.0 * r2 - 20.0 * dow) * ux;
  }
};

class ElliptBase : public StandardBaseProblem
{
public:
  typedef StandardBaseProblem super;

public:
  ElliptBase(std::string name_) : super(name_) {}

protected:
  void fillOperators()
  {
    // ===== create matrix operator =====
    Operator *matrixOperator = new Operator(getFeSpace());
    matrixOperator->addTerm(new Simple_SOT);
    prob->addMatrixOperator(matrixOperator, 0, 0);


    // ===== create rhs operator =====
    int degree = getFeSpace()->getBasisFcts()->getDegree();
    Operator *rhsOperator = new Operator(getFeSpace());
    rhsOperator->addTerm(new CoordsAtQP_ZOT(new F(degree)));
    prob->addVectorOperator(rhsOperator, 0);
  }

  void fillBoundaryCondition()
  {
    // ===== add boundary conditions =====
    prob->addDirichletBC(1, 0, 0, new G);
  }
};

// ===========================================================================
// ===== main program ========================================================
// ===========================================================================

int main(int argc, char* argv[])
{
  FUNCNAME("main");

  AMDiS::init(argc, argv);

  // ===== create and init the scalar problem ===== 
  ElliptBase ellipt("ellipt");
  ellipt.initialize(INIT_ALL);


  // === create adapt info ===
  AdaptInfo adaptInfo("ellipt->adapt", ellipt.getNumComponents());


  // === create adapt ===
  AdaptStationary adapt("ellipt->adapt", ellipt, adaptInfo);


  // ===== start adaption loop =====
  ptime start_time= microsec_clock::local_time();
  ellipt.initTimeInterface(); // fillOperators() and fillBoundaryCondition()
  adapt.adapt();
  time_duration td= microsec_clock::local_time()-start_time;

  MSG("elapsed time= %f msec\n", td.total_milliseconds()/1000.0);

  ellipt.writeFiles(&adaptInfo, true);

  AMDiS::finalize();
}
