#include "CahnHilliardNavierStokes_TwoPhase_RB.h"
#include "Views.h"
#include "SignedDistFunctors.h"
#include "PhaseFieldConvert.h"
#include "POperators.h"

using namespace AMDiS;

CahnHilliardNavierStokes_TwoPhase_RB::CahnHilliardNavierStokes_TwoPhase_RB(const std::string &name_) :
  super(name_),
  useMobility(false),
  useConservationForm(false),
  doubleWell(1),
  gamma(1.0),
  eps(0.1),
  minusEps(-0.1),
  epsInv(10.0),
  minusEpsInv(-10.0),
  epsSqr(0.01),
  minusEpsSqr(-0.01),
  laplaceType(0),
  nonLinTerm(3),
  oldTimestep(0.0),
  viscosity1(1.0),
  viscosity2(1.0),
  density1(1.0),
  density2(1.0),
  c(0.0),
  sigma(1.0),
  surfaceTension(1.0),
  reinit(NULL),
  velocity(NULL),
  fileWriter(NULL)
{
  // parameters for CH
  Parameters::get(name + "->use mobility", useMobility); // dt(c) = div(M(c)*grad(c)) + ...
  Parameters::get(name + "->gamma", gamma); // M(c) = gamma * Double-Well
  Parameters::get(name + "->epsilon", eps); // interface width
  
  // type of double well: 0= [0,1], 1= [-1,1]
  Parameters::get(name + "->double-well type", doubleWell); 

  // parameters for navier-stokes
  Initfile::get(name + "->viscosity1", viscosity1);
  Initfile::get(name + "->viscosity2", viscosity2);
  Initfile::get(name + "->density1", density1);
  Initfile::get(name + "->density2", density2);
  
  // type of laplace operator: 	0... div(nu*grad(u)), 
  //							1... div(0.5*nu*(grad(u)+grad(u)^T))
  Initfile::get(name + "->laplace operator", laplaceType); 
  
  // Parameters for CH-Coupling
  // 0... u*grad(c), 1... -div(u*c)
  Initfile::get(name + "->use conservation form", useConservationForm);

  // Parameters for NS-Coupling
  // cahn-hiliard-force: sigma*mu*grad(c)
  Initfile::get(name + "->sigma", sigma); // surface tension
  surfaceTension = sigma*3.0/(2.0*sqrt(2.0)) / eps;

    
  force.set(0.0);
  Initfile::get(name + "->force", force);
  
  // transformation of the parameters
  minusEps = -eps;
  epsInv = 1.0/eps;
  minusEpsInv = -epsInv;
  epsSqr = sqr(eps);
  minusEpsSqr = -epsSqr;
}


CahnHilliardNavierStokes_TwoPhase_RB::~CahnHilliardNavierStokes_TwoPhase_RB() 
{ FUNCNAME("CahnHilliardNavierStokes_TwoPhase_RB::~CahnHilliardNavierStokes_TwoPhase_RB()");

  if (reinit != NULL) {
    delete reinit;
    reinit = NULL;
  }
  if (velocity != NULL) {
    delete velocity;
    velocity = NULL;
  }
  delete fileWriter;
  fileWriter = NULL;
}


void CahnHilliardNavierStokes_TwoPhase_RB::initData()
{ FUNCNAME("CahnHilliardNavierStokes_TwoPhase_RB::initData()");

  dim = getMesh()->getDim();
  // create instance redistancing class
  reinit = new HL_SignedDistTraverse("reinit", dim);
  
  if (velocity == NULL)
    velocity = new DOFVector<WorldVector<double> >(getFeSpace(2), "velocity");
  fileWriter = new FileVectorWriter(name + "->velocity->output", getFeSpace(2)->getMesh(), velocity);

  super::initData();
}


void CahnHilliardNavierStokes_TwoPhase_RB::transferInitialSolution(AdaptInfo *adaptInfo)
{ FUNCNAME("CahnHilliardNavierStokes_TwoPhase_RB::transferInitialSolution()");

  calcVelocity();
  
  for (int i = 0; i < 2+dow; i++)
    prob->setExactSolution(prob->getSolution()->getDOFVector(i), i);

  fileWriter->writeFiles(adaptInfo, false);
  writeFiles(adaptInfo, false);

  // initial parameters for detecting mesh changes
  oldMeshChangeIdx= getMesh()->getChangeIndex();
}


void CahnHilliardNavierStokes_TwoPhase_RB::fillOperators()
{ FUNCNAME("CahnHilliardNavierStokes_TwoPhase_RB::fillOperators()");
  MSG("CahnHilliardNavierStokes_TwoPhase_RB::fillOperators()\n");
  
  // variable order:
  // (c, mu, u0, u1 [, u2], p)

  WorldVector<DOFVector<double>*> stage_velocity;
  WorldVector<DOFVector<double>*> un_velocity;
  for (size_t i = 0; i < dow; ++i) {
    stage_velocity[i] = prob->getStageSolution(2+i);
    un_velocity[i] = prob->getUnVec(2+i);
  }

  // dt(c) = laplace(mu) - u*grad(c)
  // -----------------------------------
  /// < dt(c) , psi >
  prob->addTimeOperator(0, 0);
  
  // div(M(c)grad(mu)), with M(c)=gamma/4*(c^2-1)^2
  Operator *opChLM = new Operator(prob->getFeSpace(0),prob->getFeSpace(1));
  if (useMobility) {
    if (doubleWell == 0) {
      // jacobian operators
      opChLM->addTerm(new VecAtQP_SOT(
	prob->getUnVec(0),
	new MobilityCH0(gamma)));
      prob->addMatrixOperator(*opChLM, 0, 1);
      
      Operator *opChLM2 = new Operator(prob->getFeSpace(0),prob->getFeSpace(0));
      opChLM2->addTerm(new VecGrad_FOT(
	prob->getUnVec(0),
	prob->getUnVec(1),
	new MobilityCH0Diff(gamma)), GRD_PSI);
      prob->addMatrixOperator(*opChLM2, 0, 0);
      
      // rhs operators
      Operator *opChLMe = new Operator(prob->getFeSpace(0),prob->getFeSpace(1));
      opChLMe->addTerm(new VecAtQP_SOT(
	prob->getStageSolution(0),
	new MobilityCH0(gamma)));
      opChLMe->setUhOld(prob->getStageSolution(1));
      prob->addVectorOperator(*opChLMe, 0, &minus1, &minus1);
    } else if (doubleWell == 1) {
      // jacobian operators
      opChLM->addTerm(new VecAtQP_SOT(
	prob->getUnVec(0),
	new MobilityCH1(gamma)));
      prob->addMatrixOperator(*opChLM, 0, 1);
      
      Operator *opChLM2 = new Operator(prob->getFeSpace(0),prob->getFeSpace(0));
      opChLM2->addTerm(new VecGrad_FOT(
	prob->getUnVec(0),
	prob->getUnVec(1),
	new MobilityCH1Diff(gamma)), GRD_PSI);
      prob->addMatrixOperator(*opChLM2, 0, 0);
      
      // rhs operators
      Operator *opChLMe = new Operator(prob->getFeSpace(0),prob->getFeSpace(1));
      opChLMe->addTerm(new VecAtQP_SOT(
	prob->getStageSolution(0),
	new MobilityCH1(gamma)));
      opChLMe->setUhOld(prob->getStageSolution(1));
      prob->addVectorOperator(*opChLMe, 0, &minus1, &minus1);
    }
  } else {
    opChLM->addTerm(new Simple_SOT(gamma));
    opChLM->setUhOld(prob->getStageSolution(1));
    prob->addVectorOperator(*opChLM, 0, &minus1, &minus1);
    prob->addMatrixOperator(*opChLM, 0, 1);                /// < phi*grad(mu) , grad(psi) >
  }
  
  /// < u_s * grad(c_s), psi >
  Operator *uGradC = new Operator(prob->getFeSpace(0), prob->getFeSpace(0));
  if (useConservationForm)
    uGradC->addTerm(new WorldVec_FOT(stage_velocity, -1.0), GRD_PSI);
  else
    uGradC->addTerm(new WorldVec_FOT(stage_velocity), GRD_PHI);  
  uGradC->setUhOld(prob->getStageSolution(0));
  prob->addVectorOperator(*uGradC, 0, &minus1, &minus1);
    
  /// < u_old * grad(c), psi >
  Operator *JuGradC1 = new Operator(prob->getFeSpace(0), prob->getFeSpace(0));
  if (useConservationForm)
    JuGradC1->addTerm(new WorldVec_FOT(un_velocity, -1.0), GRD_PSI);
  else
    JuGradC1->addTerm(new WorldVec_FOT(un_velocity), GRD_PHI);
  prob->addMatrixOperator(*JuGradC1, 0, 0);
  
  /// < u * grad(c_old), psi >
  for (size_t i = 0; i < dow; i++) {
    Operator *JuGradC2 = new Operator(prob->getFeSpace(0), prob->getFeSpace(0));
    if (useConservationForm)
      JuGradC2->addTerm(new VecAndPartialDerivative_FOT(prob->getUnVec(0), i, -1.0), GRD_PSI);
    else {
      JuGradC2->addTerm(new PartialDerivative_ZOT(prob->getUnVec(0), i));
    }
    prob->addMatrixOperator(*JuGradC2, 0, 2+i);
  }
  

  // mu + eps^2*laplace(c) - M'(c)
  // ----------------------------------------------------------------------
  /// < mu , psi >
  Operator *opChMu = new Operator(prob->getFeSpace(0),prob->getFeSpace(0));
  opChMu->addZeroOrderTerm(new Simple_ZOT);
  opChMu->setUhOld(prob->getStageSolution(1));
  prob->addVectorOperator(*opChMu, 1);
  prob->addMatrixOperator(*opChMu, 1, 1, &minus1, &minus1); 
  
  /// < -eps^2*grad(c) , grad(psi) >
  Operator *opChLC = new Operator(prob->getFeSpace(1),prob->getFeSpace(0));
  opChLC->addSecondOrderTerm(new Simple_SOT(minusEpsSqr));
  opChLC->setUhOld(prob->getStageSolution(0));
  prob->addVectorOperator(*opChLC, 1);
  prob->addMatrixOperator(*opChLC, 1, 0, &minus1, &minus1);   
  
  /// < -M'(c) , psi >
  if (doubleWell == 0) {
    // jacobian operators
    Operator *opChMImpl = new Operator(prob->getFeSpace(1),prob->getFeSpace(0));
    opChMImpl->addZeroOrderTerm(new VecAtQP_ZOT(
      prob->getUnVec(0),
      new DoubleWell0Diff(-1.0))); // < (3c^2-3c+0.5)*c' , psi >
    prob->addMatrixOperator(*opChMImpl, 1, 0, &minus1, &minus1);   
    // rhs operators
    Operator *opChMExpl = new Operator(prob->getFeSpace(1),prob->getFeSpace(0));
    opChMExpl->addZeroOrderTerm(new VecAtQP_ZOT(
      prob->getStageSolution(0),
      new DoubleWell0(-1.0))); // < (c^3-3/2*c^2+0.5) , psi >
    prob->addVectorOperator(*opChMExpl, 1);   
  }
  else if (doubleWell == 1) {
    // jacobian operators
    Operator *opChMImpl = new Operator(prob->getFeSpace(1),prob->getFeSpace(0));
    opChMImpl->addZeroOrderTerm(new VecAtQP_ZOT(
      prob->getUnVec(0),
      new DoubleWell1Diff(-1.0))); // < (3c^2-1)*c' , psi >
    prob->addMatrixOperator(*opChMImpl, 1, 0, &minus1, &minus1);   
    // rhs operators
    Operator *opChMExpl = new Operator(prob->getFeSpace(1),prob->getFeSpace(0));
    opChMExpl->addZeroOrderTerm(new VecAtQP_ZOT(
      prob->getStageSolution(0),
      new DoubleWell1(-1.0))); // < (c^3-c) , psi >
    prob->addVectorOperator(*opChMExpl, 1);   
  }

  // Navier-Stokes part
  // -----------------------------------------------------------------------------------
  bool variableDensity = abs(density1 - density2) > DBL_TOL;
  int shift = 2;
  if (variableDensity)
    shift = 3+dow;
  
  for (size_t i = 0; i < dow; ++i) {
    /// < d_t(u_i) , psi >
    prob->addTimeOperator(shift+i, 2+i, (variableDensity ? 1.0 : density1));
    
    if (variableDensity) {
      /// < z , psi >
      Operator *opZ = new Operator(getFeSpace(shift+i), getFeSpace(shift+i));
      opZ->addTerm(new Simple_ZOT);
      prob->addMatrixOperator(*opZ, shift+i, shift+i, &minus1, &minus1);
      opZ->setUhOld(prob->getStageSolution(shift+i));
      prob->addVectorOperator(*opZ, shift+i);
    }
	
    /// < u_s*grad(u_s_i) , psi >
    Operator *opUGradU = new Operator(getFeSpace(shift+i), getFeSpace(2+i));
    opUGradU->addTerm(new WorldVec_FOT(stage_velocity, (variableDensity ? -1.0 : -density1)), GRD_PHI);
    opUGradU->setUhOld(stage_velocity[i]);
    prob->addVectorOperator(*opUGradU, shift+i);

    /// < u_old*grad(u_i) , psi >
    Operator *opUGradV = new Operator(getFeSpace(shift+i), getFeSpace(2+i));
    opUGradV->addTerm(new WorldVec_FOT(un_velocity, (variableDensity ? -1.0 : -density1)), GRD_PHI);
    prob->addMatrixOperator(*opUGradV, shift+i, 2+i, &minus1, &minus1);

    /// < u*grad(u_old_i) , psi >
    for (size_t j = 0; j < dow; ++j) {
      Operator *opVGradU = new Operator(getFeSpace(shift+i), getFeSpace(2+j));
      opVGradU->addTerm(new PartialDerivative_ZOT(un_velocity[i], j, (variableDensity ? -1.0 : -density1)));
      prob->addMatrixOperator(*opVGradU, shift+i, 2+j, &minus1, &minus1);
    }
  }
  
  for (size_t i = 0; i < dow; ++i) {

    /// Diffusion-Operator (including Stress-Tensor for space-dependent viscosity)
    addStressTerm(i);
  
    /// external force, i.e. gravitational force
    if (norm(force) > DBL_TOL) {
      Operator *opForce = new Operator(getFeSpace(2+i), getFeSpace(2+i));
      opForce->addTerm(new Simple_ZOT(force[i]));
      prob->addVectorOperator(*opForce, 2+i);
    }
    
    // forces by Cahn-Hilliard terms
    // -----------------------------
    if (surfaceTension > DBL_TOL) {
      /// < mu_old * grad(c) , psi >
      Operator *JmuGradC = new Operator(getFeSpace(2+i),getFeSpace(0));
      JmuGradC->addTerm(new VecAndPartialDerivative_FOT(
	prob->getUnVec(1), i, -1.0), GRD_PHI);
      prob->addMatrixOperator(*JmuGradC, 2+i, 0, &surfaceTension, &surfaceTension);
      
	  /// < mu * grad(c_old) , psi >
      Operator *JmuGradC2 = new Operator(getFeSpace(2+i),getFeSpace(1));
      JmuGradC2->addTerm(new PartialDerivative_ZOT(
	prob->getUnVec(0), i, -1.0));
      prob->addMatrixOperator(*JmuGradC2, 2+i, 1, &surfaceTension, &surfaceTension);
	  
	  /// < mu_s * grad(c_s) , psi >
      Operator *opMuGradC = new Operator(getFeSpace(2+i),getFeSpace(1));
      opMuGradC->addTerm(new VecAndPartialDerivative_ZOT(
	    prob->getStageSolution(1),
	prob->getStageSolution(0), i));
      prob->addVectorOperator(*opMuGradC, 2+i, &surfaceTension, &surfaceTension);
    }
	
    // ---------------------------------------
    if (variableDensity) {
      /// < rho(c_s)*z_s , psi > 
      Operator *opDensityExpl = new Operator(prob->getFeSpace(2+i),prob->getFeSpace(shift+1));
      opDensityExpl->addZeroOrderTerm(new VecAtQP_ZOT(
	prob->getStageSolution(0),
	new LinearInterpolation(density1, density2, -1.0))); 
      opDensityExpl->setUhOld(prob->getStageSolution(shift+i));
      prob->addVectorOperator(*opDensityExpl, 2+i);   
	  
      /// < rho(c_old)*z , psi > 
      Operator *opDensityImpl = new Operator(prob->getFeSpace(1),prob->getFeSpace(shift+i));
      opDensityImpl->addZeroOrderTerm(new VecAtQP_ZOT(
	prob->getUnVec(0),
	new LinearInterpolation(density1, density2, -1.0))); 
      prob->addMatrixOperator(*opDensityImpl, 2+i, shift+i, &minus1, &minus1);   
	  
      /// < rho'(c_old)*z_old * c , psi > 
      Operator *opDensityImpl2 = new Operator(prob->getFeSpace(1),prob->getFeSpace(0));
      opDensityImpl2->addZeroOrderTerm(new Vec2AtQP_ZOT(
	prob->getUnVec(0),
	prob->getUnVec(shift+i),
	new LinearInterpolationDiffFactor(density1, density2, -1.0))); 
      prob->addMatrixOperator(*opDensityImpl2, 2+i, 0, &minus1, &minus1);   
    }
  }

  // div(u) = 0
  // ----------
  for (size_t i = 0; i < dow; ++i) {
    /// < d_i(u_i) , psi >
    Operator *opDivU = new Operator(getFeSpace(2+dow),getFeSpace(2+i));
    opDivU->addTerm(new PartialDerivative_FOT(i), GRD_PHI);
    prob->addMatrixOperator(*opDivU, 2+dow, 2+i, &minus1, &minus1);
    /// < d_i(u_s_i) , psi >
    opDivU->setUhOld(stage_velocity[i]);
    prob->addVectorOperator(*opDivU, 2+dow);
  }
}


// void CahnHilliardNavierStokes_TwoPhase_RB::addNonlinearTerm(int i)
// { FUNCNAME("CahnHilliardNavierStokes_TwoPhase_RB::addNonlinearTerm()");
//   
//   WorldVector<DOFVector<double>*> stage_velocity;
//   WorldVector<DOFVector<double>*> un_velocity;
//   for (size_t j = 0; j < dow; ++j) {
//     stage_velocity[j] = prob->getStageSolution(2+j);
//     un_velocity[j] = prob->getUnVec(2+j);
//   }
//   
//   /// < rho(c_s)*u_s*grad(u_s_i) , psi >
//   Operator *opUGradU = new Operator(getFeSpace(2+i), getFeSpace(2+i));
//   opUGradU->addTerm(new VecAndWorldVec_FOT(
// 	prob->getStageSolution(0),
// 	stage_velocity,
// 	new LinearInterpolation(density1, density2, -1.0)), GRD_PHI);
//   opUGradU->setUhOld(stage_velocity[i]);
//   prob->addVectorOperator(*opUGradU, 2+i);
// 
//   /// < rho(c_old)*u_old*grad(u_i) , psi >
//   Operator *opUGradV = new Operator(getFeSpace(2+i), getFeSpace(2+i));
//   opUGradV->addTerm(new VecAndWorldVec_FOT(
// 	prob->getUnVec(0),
// 	un_velocity,
// 	new LinearInterpolation(density1, density2, -1.0)), GRD_PHI);
//   prob->addMatrixOperator(*opUGradV, 2+i, 2+i, &minus1, &minus1);
// 
//   /// < c * rho'(c_old)*u_old*grad(u_old_i) , psi >
//   for (size_t j = 0; j < dow; ++j) {
// 	Operator *opVGradU = new Operator(getFeSpace(2+i), getFeSpace(2+j));
// 	opVGradU->addTerm(new Vec2AndPartialDerivative_ZOT(
// 	  prob->getUnVec(0),
// 	  un_velocity[i],
// 	  un_velocity[j],
// 	  j, 
// 	  new LinearInterpolationDiffFactor(density1, density2, -1.0)));
// 	prob->addMatrixOperator(*opVGradU, 2+i, 0, &minus1, &minus1);
//   }
// }


void CahnHilliardNavierStokes_TwoPhase_RB::addStressTerm(int i)
{ FUNCNAME("CahnHilliardNavierStokes_TwoPhase_RB::addStressTerm()");

    /// < nu*[grad(u)+grad(u)^t] , grad(psi) >
    if (laplaceType == 1) {
      for (unsigned j = 0; j < dow; ++j) {	
	/// < nu(c_s)*d_j(u_s_i) , d_i(psi) >	
        Operator *opLaplaceUijExpl = new Operator(getFeSpace(2+i), getFeSpace(2+j));
        opLaplaceUijExpl->addTerm(new VecAtQP_IJ_SOT(
	  prob->getStageSolution(0),
	  new LinearInterpolation(viscosity1, viscosity2),
	  1-i, 1-j));
	opLaplaceUijExpl->setUhOld(prob->getStageSolution(2+j));
	prob->addVectorOperator(*opLaplaceUijExpl, 2+i, &minus1, &minus1);
		
	/// < nu(c_old)*d_j(u_i) , d_i(psi) >
        Operator *opLaplaceUijImpl = new Operator(getFeSpace(2+i), getFeSpace(2+j));
        opLaplaceUijImpl->addTerm(new VecAtQP_IJ_SOT(
	  prob->getUnVec(0),
	  new LinearInterpolation(viscosity1, viscosity2),
	  1-i, 1-j));
        prob->addMatrixOperator(*opLaplaceUijImpl, 2+i, 2+j);
		
	/// < c * nu'(c_old)*d_j(u_old_i) , d_i(psi) >
        Operator *opLaplaceUijImpl2 = new Operator(getFeSpace(2+i), getFeSpace(2+j));
        opLaplaceUijImpl2->addTerm(new VecAndPartialDerivativeIJ_FOT(
	  prob->getUnVec(0),
	  prob->getUnVec(2+j),
	  1-i, 1-j,
	  new LinearInterpolationDiffFactor(viscosity1, viscosity2)), GRD_PSI);
        prob->addMatrixOperator(*opLaplaceUijImpl2, 2+i, 0);
      }
    }
    
    /// < nu*grad(u'_i) , grad(psi) >
    Operator *opLaplaceUExpl = new Operator(getFeSpace(2+i), getFeSpace(2+i));
    opLaplaceUExpl->addTerm(new VecAtQP_SOT(
      prob->getStageSolution(0),
      new LinearInterpolation(viscosity1, viscosity2)));
    opLaplaceUExpl->setUhOld(prob->getStageSolution(2+i));
    prob->addVectorOperator(*opLaplaceUExpl, 2+i, &minus1, &minus1);
	
    Operator *opLaplaceUImpl = new Operator(getFeSpace(2+i), getFeSpace(2+i));
    opLaplaceUImpl->addTerm(new VecAtQP_SOT(
      prob->getUnVec(0),
      new LinearInterpolation(viscosity1, viscosity2)));
    prob->addMatrixOperator(*opLaplaceUImpl, 2+i, 2+i);
	
    Operator *opLaplaceUImpl2 = new Operator(getFeSpace(2+i), getFeSpace(2+i));
    opLaplaceUImpl2->addTerm(new VecGrad_FOT(
      prob->getUnVec(0),
      prob->getUnVec(2+i),
      new LinearInterpolationDiffVector(viscosity1, viscosity2)), GRD_PSI);
    prob->addMatrixOperator(*opLaplaceUImpl2, 2+i, 0);
  
    /// < p , d_i(psi) >
    Operator *opGradP = new Operator(getFeSpace(2+i),getFeSpace(2+dow));
    opGradP->addTerm(new PartialDerivative_FOT(i), GRD_PSI);
    opGradP->setUhOld(prob->getStageSolution(2+dow));
    prob->addVectorOperator(*opGradP, 2+i);
    prob->addMatrixOperator(*opGradP, 2+i, 2+dow, &minus1, &minus1);
}

void CahnHilliardNavierStokes_TwoPhase_RB::closeTimestep(AdaptInfo *adaptInfo)
{ FUNCNAME("CahnHilliardNavierStokes_TwoPhase_RB::closeTimestep()");

  calcVelocity();
  fileWriter->writeFiles(adaptInfo, false);
  writeFiles(adaptInfo, false);
}
