#include "NavierStokes_TH_MultiPhase_RB.h"

using namespace std;
using namespace AMDiS;

NavierStokes_TH_MultiPhase_RB::NavierStokes_TH_MultiPhase_RB(const std::string &name_) :
  super(name_),
  multiPhase(NULL),
  densityPhase(NULL),
  viscosityPhase(NULL),
  viscosity1(-1.0),
  viscosity2(1.0),
  density1(1.0),
  density2(1.0)
{
  Initfile::get(name + "->viscosity1", viscosity1); // viscosity of fluid 1
  Initfile::get(name + "->viscosity2", viscosity2); // viscosity of fluid 2
  Initfile::get(name + "->density1", density1); // density of fluid 1
  Initfile::get(name + "->density2", density2); // density of fluid 2

  if (viscosity1 <= 0.0)
    viscosity1 = viscosity;
};


NavierStokes_TH_MultiPhase_RB::~NavierStokes_TH_MultiPhase_RB()
{ FUNCNAME("NavierStokes_TH_MultiPhase_RB::~NavierStokes_TH_MultiPhase_RB()");

  delete multiPhase;
  delete densityPhase;
  delete viscosityPhase;
};


void NavierStokes_TH_MultiPhase_RB::initData()
{ FUNCNAME("NavierStokes_TH_MultiPhase_RB::initTimeInterface()");

  multiPhase = new DOFVector<double>(prob->getFeSpace(0), "multiPhase");
  densityPhase = new DOFVector<double>(prob->getFeSpace(0), "densityPhase");
  viscosityPhase = new DOFVector<double>(prob->getFeSpace(0), "viscosityPhase");

  multiPhase->set(1.0);
  densityPhase->set(density1);
  viscosityPhase->set(viscosity1);

  super::initData();
};

void NavierStokes_TH_MultiPhase_RB::initTimestep(AdaptInfo *adaptInfo)
{ FUNCNAME("NavierStokes_TH_MultiPhase_RB::initTimestep()");

  super::initTimestep(adaptInfo);

  transformDOF(multiPhase, densityPhase, new LinearInterpolation(density1, density2));
  transformDOF(multiPhase, viscosityPhase, new LinearInterpolation(viscosity1, viscosity2));
};


void NavierStokes_TH_MultiPhase_RB::fillOperators()
{ FUNCNAME("NavierStokes_TH_MultiPhase_RB::fillOperators()");

  WorldVector<DOFVector<double>* > vel;
  for (unsigned i = 0; i < dim; i++)
    vel[i]= prob->getSolution()->getDOFVector(i);

  WorldVector<DOFVector<double>*> stage_velocity;
  WorldVector<DOFVector<double>*> un_velocity;
  for (unsigned i = 0; i < dow; ++i) {
    stage_velocity[i] = prob->getStageSolution(i);
    un_velocity[i] = prob->getUnVec(i);
  }
  
  // fill operators for prob
  for (unsigned i = 0; i < dim; ++i) {
    /// < d_t(u_i) , psi >
    Operator *opTime = new Operator(getFeSpace(i), getFeSpace(i));
    opTime->addZeroOrderTerm(new VecAtQP_ZOT(densityPhase, NULL));
    prob->addMatrixOperator(opTime, i, i, prob->getInvTauGamma(), prob->getInvTauGamma());

    Operator *opTimeOld = new Operator(getFeSpace(i));
    opTimeOld->addZeroOrderTerm(new Vec2AtQP_ZOT(densityPhase, prob->getTimeRhsVec(i), NULL));
    prob->addVectorOperator(opTimeOld, i, &minus1, &minus1);

    /// < u*grad(u_i) , psi >
    Operator *opUGradU = new Operator(getFeSpace(i), getFeSpace(i));
    opUGradU->addTerm(new WorldVecPhase_FOT(densityPhase, stage_velocity, -1.0), GRD_PHI);
    opUGradU->setUhOld(stage_velocity[i]);
    prob->addVectorOperator(*opUGradU, i);

    Operator *opUGradV = new Operator(getFeSpace(i), getFeSpace(i));
    opUGradV->addTerm(new WorldVecPhase_FOT(densityPhase, un_velocity, -1.0), GRD_PHI);
    prob->addMatrixOperator(*opUGradV, i, i, &minus1, &minus1);

    for (unsigned j = 0; j < dow; ++j) {
      Operator *opVGradU = new Operator(getFeSpace(i), getFeSpace(j));
      opVGradU->addTerm(new VecAndPartialDerivative_ZOT(densityPhase, un_velocity[i], j, -1.0));
      prob->addMatrixOperator(*opVGradU, i, j, &minus1, &minus1);
    }

    /// Diffusion-Operator (including Stress-Tensor for space-dependent viscosity
    addLaplaceTerm(i);

    /// < p , d_i(psi) >
    Operator *opGradP = new Operator(getFeSpace(i),getFeSpace(dow));
    opGradP->addTerm(new PartialDerivative_FOT(i, -1.0), GRD_PSI);
    opGradP->setUhOld(prob->getStageSolution(dow));
    prob->addVectorOperator(*opGradP, i);
    prob->addMatrixOperator(*opGradP, i, dow, &minus1, &minus1);
  
    /// external force, i.e. gravitational force
    if (norm(force) > DBL_TOL) {
      Operator *opForce = new Operator(getFeSpace(i), getFeSpace(i));
      opForce->addZeroOrderTerm(new Phase_ZOT(densityPhase, force[i]));
      prob->addVectorOperator(*opForce, i);
    }
  }

  /// div(u) = 0
  for (unsigned i = 0; i < dow; ++i) {
    /// < d_i(u'_i) , psi >
    Operator *opDivU = new Operator(getFeSpace(dow),getFeSpace(i));
    opDivU->addTerm(new PartialDerivative_FOT(i), GRD_PHI);
    opDivU->setUhOld(stage_velocity[i]);
    prob->addVectorOperator(*opDivU, dow);
    prob->addMatrixOperator(*opDivU, dow, i, &minus1, &minus1);
  }
};


void NavierStokes_TH_MultiPhase_RB::addLaplaceTerm(int i)
{ FUNCNAME("NavierStokes_TH_MultiPhase_RB::addLaplaceTerm()");

  /// < alpha*[grad(u)+grad(u)^t] , grad(psi) >
  if (laplaceType == 1) {
    for (unsigned j = 0; j < dim; ++j) {
      Operator *opLaplaceUi1 = new Operator(getFeSpace(i), getFeSpace(j));
      opLaplaceUi1->addTerm(new VecAtQP_IJ_SOT(
        viscosityPhase, NULL, 1-i, 1-j));
      opLaplaceUi1->setUhOld(prob->getStageSolution(j));
      prob->addVectorOperator(*opLaplaceUi1, i, &minus1, &minus1);
      prob->addMatrixOperator(*opLaplaceUi1, i, j);
    }
  }
  
  /// < alpha*grad(u'_i) , grad(psi) >
  Operator *opLaplaceUi = new Operator(getFeSpace(i), getFeSpace(i));
    opLaplaceUi->addTerm(new VecAtQP_SOT(viscosityPhase, NULL));
  opLaplaceUi->setUhOld(prob->getStageSolution(i));
  prob->addVectorOperator(*opLaplaceUi, i, &minus1, &minus1);
  prob->addMatrixOperator(*opLaplaceUi, i, i);
};
